/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.container.ri.client;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import org.jdtaus.core.container.Container;
import org.jdtaus.core.container.ContainerError;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.ContextFactory;
import org.jdtaus.core.container.ContextInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Specification;

/**
 * {@code Container} reference implementation.
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: DefaultContainer.java 4590 2008-03-10 01:40:58Z schulte2005 $
 *
 * @see org.jdtaus.core.container.ContainerFactory
 */
public class DefaultContainer implements Container
{
    //--Constants---------------------------------------------------------------

    /** Empty Class-Array. */
    private static final Class[] EMPTY = {};

    /** Implementation Class-Array. */
    private static final Class[] IMPL_CTOR = { Implementation.class };

    /** Dependency Class-Array. */
    private static final Class[] DEP_CTOR = { Dependency.class };

    //---------------------------------------------------------------Constants--
    //--Container---------------------------------------------------------------

    /** Maps implementation identifiers to singleton instances. */
    private final Map singletons = new HashMap( 100 );

    /** Mutex used to detect cyclic instantiations. */
    private final Object cycle = new Object();

    public final Object getImplementation(
        final Class specification, final String implementationName )
    {
        Object instance = null;
        final Implementation impl;
        final Specification spec;

        if ( specification == null )
        {
            throw new NullPointerException( "specification" );
        }
        if ( implementationName == null )
        {
            throw new NullPointerException( "implementationName" );
        }

        // Find the implementation.
        spec = ModelFactory.getModel().getModules().
            getSpecification( specification.getName() );

        impl = spec.getImplementation( implementationName );

        if ( spec.isSingleton() )
        {
            // Return the singleton instance for impl.
            instance = this.singletons.get( impl.getIdentifier() );
            if ( instance == null || instance instanceof Implementation )
            {
                synchronized ( this.cycle )
                {
                    instance = this.singletons.get( impl.getIdentifier() );
                    if ( instance == null )
                    {
                        this.singletons.put( impl.getIdentifier(), impl );
                        try
                        {
                            instance = this.instantiateImplementation( impl );
                        }
                        catch ( org.jdtaus.core.container.InstantiationException e )
                        {
                            this.singletons.remove( impl.getIdentifier() );
                            throw e;
                        }
                        this.singletons.put( impl.getIdentifier(), instance );
                    }
                    else if ( instance instanceof Implementation )
                    {
                        throw new DependencyCycleException(
                            ( Implementation ) instance, impl );

                    }
                }
            }
        }
        else
        {
            // Create a new instance.
            instance = this.instantiateImplementation( impl );
        }

        this.initializeContext( instance );
        return instance;
    }

    public final Object getDependency(
        final Class implementation, final String dependencyName )
    {
        final Dependency dep;
        final Implementation impl;
        Object instance = null;

        if ( implementation == null )
        {
            throw new NullPointerException( "implementation" );
        }
        if ( dependencyName == null )
        {
            throw new NullPointerException( "dependencyName" );
        }

        impl = ModelFactory.getModel().getModules().
            getImplementation( implementation.getName() );

        dep = impl.getDependencies().getDependency( dependencyName );

        if ( dep.getSpecification().isSingleton() )
        {
            // Return the singleton instance for impl.
            instance = this.singletons.get( dep.getImplementation().
                                            getIdentifier() );

            if ( instance == null || instance instanceof Implementation )
            {
                synchronized ( this.cycle )
                {
                    instance = this.singletons.get( dep.getImplementation().
                                                    getIdentifier() );

                    if ( instance == null )
                    {
                        this.singletons.put( dep.getImplementation().
                                             getIdentifier(),
                                             dep.getImplementation() );

                        try
                        {
                            instance = this.instantiateDependency( dep );
                        }
                        catch ( org.jdtaus.core.container.InstantiationException e )
                        {
                            this.singletons.remove( dep.getImplementation().
                                                    getIdentifier() );

                            throw e;
                        }

                        this.singletons.put( dep.getImplementation().
                                             getIdentifier(), instance );

                    }
                    else if ( instance instanceof Implementation )
                    {
                        throw new DependencyCycleException(
                            ( Implementation ) instance, impl );

                    }
                }
            }
        }
        else
        {
            instance = this.instantiateDependency( dep );
        }

        this.initializeContext( instance );
        return instance;
    }

    //---------------------------------------------------------------Container--
    //--DefaultContainer--------------------------------------------------------

    /**
     * Gets the classloader used for loading classes.
     * <p>The reference implementation will use the current thread's context
     * classloader and will fall back to the system classloader if the
     * current thread has no context classloader set.</p>
     *
     * @return the classloader to be used for loading classes.
     *
     * @throws ContainerError if no classloader is available.
     */
    protected ClassLoader getClassLoader()
    {
        ClassLoader classLoader = Thread.currentThread().
            getContextClassLoader();

        if ( classLoader == null )
        {
            classLoader = ClassLoader.getSystemClassLoader();
        }

        if ( classLoader == null )
        {
            throw new ContainerError(
                DefaultContainerBundle.getInstance().
                getMissingClassloaderErrorText( Locale.getDefault() ) );

        }

        return classLoader;
    }

    /**
     * Creates a new instance of an implementation.
     * <p>The reference implementation will load the class corresponding to
     * the implementation by using the classloader returned by method
     * {@link #getClassLoader()} and will then create a new instance of that
     * class by calling the constructor taking a {@code Implementation} as
     * a parameter.</p>
     *
     * @param impl the implementation to return a new instance for.
     *
     * @return a new instance of the class corresponding to {@code impl}.
     *
     * @throws NullPointerException if {@code impl} is {@code null}.
     * @throws org.jdtaus.core.container.InstantiationException
     * if no instance can be created.
     * @throws ContainerError if no classloader is available.
     */
    protected Object instantiateImplementation( final Implementation impl )
    {
        final Object instance;
        final Constructor ctor;
        final Class clazz;

        if ( impl == null )
        {
            throw new NullPointerException( "impl" );
        }

        try
        {
            clazz = this.getClassLoader().loadClass( impl.getIdentifier() );
            ctor = clazz.getDeclaredConstructor( DefaultContainer.IMPL_CTOR );
            ctor.setAccessible( true );
            instance = ctor.newInstance( new Object[] { impl } );
            this.initialize( instance );
            return instance;
        }
        catch ( ClassNotFoundException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
        catch ( NoSuchMethodException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
        catch ( IllegalArgumentException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
        catch ( InvocationTargetException e )
        {
            final Throwable targetException = e.getTargetException();

            if ( targetException instanceof Error )
            {
                throw ( Error ) targetException;
            }
            else if ( targetException instanceof RuntimeException )
            {
                throw ( RuntimeException ) targetException;
            }
            else
            {
                throw new org.jdtaus.core.container.InstantiationException(
                    impl.getIdentifier(), targetException == null
                    ? e
                    : targetException );

            }
        }
        catch ( IllegalAccessException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
        catch ( InstantiationException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
        catch ( Throwable e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                impl.getIdentifier(), e );

        }
    }

    /**
     * Creates a new instance of a dependency.
     * <p>The reference implementation will load the class corresponding to
     * the dependency by using the classloader returned by method
     * {@link #getClassLoader()} and will then create a new instance of that
     * class by calling the constructor taking a {@code Dependency} as
     * a parameter.</p>
     *
     * @param dep the dependency to return a new instance for.
     *
     * @return a new instance of the class corresponding to {@code dep}.
     *
     * @throws NullPointerException if {@code dep} is {@code null}.
     * @throws org.jdtaus.core.container.InstantiationException
     * if no instance can be created.
     * @throws ContainerError if no classloader is available.
     */
    protected Object instantiateDependency( final Dependency dep )
    {
        final Object instance;
        final Constructor ctor;
        final Class clazz;

        if ( dep == null )
        {
            throw new NullPointerException( "dep" );
        }

        try
        {
            clazz = this.getClassLoader().loadClass(
                dep.getImplementation().getIdentifier() );

            ctor = clazz.getDeclaredConstructor( DefaultContainer.DEP_CTOR );
            ctor.setAccessible( true );
            instance = ctor.newInstance( new Object[] { dep } );
            this.initialize( instance );
            return instance;
        }
        catch ( ClassNotFoundException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
        catch ( NoSuchMethodException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
        catch ( IllegalArgumentException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
        catch ( InvocationTargetException e )
        {
            final Throwable targetException = e.getTargetException();

            if ( targetException instanceof Error )
            {
                throw ( Error ) targetException;
            }
            else if ( targetException instanceof RuntimeException )
            {
                throw ( RuntimeException ) targetException;
            }
            else
            {
                throw new org.jdtaus.core.container.InstantiationException(
                    dep.getImplementation().getIdentifier(),
                    targetException == null
                    ? e
                    : targetException );

            }
        }
        catch ( IllegalAccessException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
        catch ( InstantiationException e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
        catch ( Throwable e )
        {
            throw new org.jdtaus.core.container.InstantiationException(
                dep.getImplementation().getIdentifier(), e );

        }
    }

    /**
     * Performs custom initialization of an instance.
     *
     * @param instance the instance to perform custom initialization with.
     *
     * @throws NullPointerException if {@code instance} is {@code null}.
     */
    protected void initialize( final Object instance )
    {
        if ( instance == null )
        {
            throw new NullPointerException( "instance" );
        }

        if ( instance instanceof ContainerInitializer )
        {
            ( ( ContainerInitializer ) instance ).initialize();
        }
    }

    /**
     * Performs context initialization of an instance.
     *
     * @param instance the instance to perform context initialization with.
     *
     * @throws NullPointerException if {@code instance} is {@code null}.
     * @throws org.jdtaus.core.container.ContextError for unrecoverable context
     * errors.
     */
    protected void initializeContext( final Object instance )
    {
        if ( instance == null )
        {
            throw new NullPointerException( "instance" );
        }

        if ( instance instanceof ContextInitializer &&
            !( ( ContextInitializer ) instance ).isInitialized(
            ContextFactory.getContext() ) )
        {
            ( ( ContextInitializer ) instance ).initialize(
                ContextFactory.getContext() );

        }
    }

    //--------------------------------------------------------DefaultContainer--
}
